#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest import mock
from unittest.mock import MagicMock

import pytest

try:
    import importlib.util

    if not importlib.util.find_spec("airflow.sdk.bases.hook"):
        raise ImportError

    BASEHOOK_PATCH_PATH = "airflow.sdk.bases.hook.BaseHook"
except ImportError:
    BASEHOOK_PATCH_PATH = "airflow.hooks.base.BaseHook"
from airflow.models import Connection
from airflow.providers.common.compat.openlineage.facet import (
    Dataset,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
    SQLJobFacet,
)
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.providers.google.cloud.operators.cloud_sql import (
    CloudSQLCloneInstanceOperator,
    CloudSQLCreateInstanceDatabaseOperator,
    CloudSQLCreateInstanceOperator,
    CloudSQLDeleteInstanceDatabaseOperator,
    CloudSQLDeleteInstanceOperator,
    CloudSQLExecuteQueryOperator,
    CloudSQLExportInstanceOperator,
    CloudSQLImportInstanceOperator,
    CloudSQLInstancePatchOperator,
    CloudSQLPatchInstanceDatabaseOperator,
)
from airflow.providers.google.cloud.triggers.cloud_sql import CloudSQLExportTrigger
from airflow.providers.google.common.consts import GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME

PROJECT_ID = os.environ.get("PROJECT_ID", "project-id")
INSTANCE_NAME = os.environ.get("INSTANCE_NAME", "test-name")
DB_NAME = os.environ.get("DB_NAME", "db1")

CREATE_BODY = {
    "name": INSTANCE_NAME,
    "settings": {
        "tier": "db-n1-standard-1",
        "backupConfiguration": {
            "binaryLogEnabled": True,
            "enabled": True,
            "replicationLogArchivingEnabled": True,
            "startTime": "05:00",
        },
        "activationPolicy": "ALWAYS",
        "authorizedGaeApplications": [],
        "crashSafeReplicationEnabled": True,
        "dataDiskSizeGb": 30,
        "dataDiskType": "PD_SSD",
        "databaseFlags": [],
        "ipConfiguration": {
            "ipv4Enabled": True,
            "authorizedNetworks": [
                {
                    "value": "192.168.100.0/24",
                    "name": "network1",
                    "expirationTime": "2012-11-15T16:19:00.094Z",
                },
            ],
            "privateNetwork": "/vpc/resource/link",
            "requireSsl": True,
        },
        "locationPreference": {
            "zone": "europe-west4-a",
            "followGaeApplication": "/app/engine/application/to/follow",
        },
        "maintenanceWindow": {"hour": 5, "day": 7, "updateTrack": "canary"},
        "pricingPlan": "PER_USE",
        "replicationType": "ASYNCHRONOUS",
        "storageAutoResize": False,
        "storageAutoResizeLimit": 0,
        "userLabels": {"my-key": "my-value"},
    },
    "databaseVersion": "MYSQL_5_7",
    "failoverReplica": {"name": "replica-1"},
    "masterInstanceName": "master-instance-1",
    "onPremisesConfiguration": {},
    "region": "europe-west4",
    "replicaConfiguration": {
        "mysqlReplicaConfiguration": {
            "caCertificate": "cert-pem",
            "clientCertificate": "cert-pem",
            "clientKey": "cert-pem",
            "connectRetryInterval": 30,
            "dumpFilePath": "/path/to/dump",
            "masterHeartbeatPeriod": 100,
            "password": "secret_pass",
            "sslCipher": "list-of-ciphers",
            "username": "user",
            "verifyServerCertificate": True,
        },
    },
}
PATCH_BODY = {
    "name": INSTANCE_NAME,
    "settings": {"tier": "db-n1-standard-2", "dataDiskType": "PD_HDD"},
    "region": "europe-west4",
}
DATABASE_INSERT_BODY = {
    "name": DB_NAME,  # The name of the database in the Cloud SQL instance.
    # This does not include the project ID or instance name.
    "project": PROJECT_ID,  # The project ID of the project containing the Cloud SQL
    # database. The Google apps domain is prefixed if
    # applicable.
    "instance": INSTANCE_NAME,  # The name of the Cloud SQL instance.
    # This does not include the project ID.
}
DATABASE_PATCH_BODY = {"charset": "utf16", "collation": "utf16_general_ci"}
EXPORT_BODY = {
    "exportContext": {
        "fileType": "CSV",
        "uri": "gs://bucketName/fileName",
        "databases": [],
        "sqlExportOptions": {
            "tables": ["table1", "table2"],
            "schemaOnly": False,
            "mysqlExportOptions": {
                "masterData": 1,
            },
        },
        "csvExportOptions": {
            "selectQuery": "SELECT * FROM TABLE",
            "escapeCharacter": "e",
            "quoteCharacter": "q",
            "fieldsTerminatedBy": "f",
            "linesTerminatedBy": "l",
        },
        "offload": True,
    }
}
IMPORT_BODY = {
    "importContext": {
        "fileType": "CSV",
        "uri": "gs://bucketName/fileName",
        "database": "db1",
        "importUser": "",
        "csvImportOptions": {"table": "my_table", "columns": ["col1", "col2"]},
    }
}


class TestCloudSql:
    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_create(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = False
        mock_hook.return_value.create_instance.return_value = True
        op = CloudSQLCreateInstanceOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=CREATE_BODY, task_id="id"
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            project_id=PROJECT_ID, body=CREATE_BODY
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_create_missing_project_id(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = False
        mock_hook.return_value.create_instance.return_value = True
        op = CloudSQLCreateInstanceOperator(instance=INSTANCE_NAME, body=CREATE_BODY, task_id="id")
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(project_id=None, body=CREATE_BODY)

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_create_idempotent(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = True
        mock_hook.return_value.create_instance.return_value = True
        op = CloudSQLCreateInstanceOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=CREATE_BODY, task_id="id"
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_create_should_throw_ex_when_empty_project_id(self, mock_hook):
        with pytest.raises(AirflowException, match="The required parameter 'project_id' is empty"):
            CloudSQLCreateInstanceOperator(
                project_id="", body=CREATE_BODY, instance=INSTANCE_NAME, task_id="id"
            )

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_create_should_throw_ex_when_empty_body(self, mock_hook):
        with pytest.raises(AirflowException, match="The required parameter 'body' is empty"):
            CloudSQLCreateInstanceOperator(
                project_id=PROJECT_ID, body={}, instance=INSTANCE_NAME, task_id="id"
            )

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_create_should_throw_ex_when_empty_instance(self, mock_hook):
        with pytest.raises(AirflowException, match="The required parameter 'instance' is empty"):
            CloudSQLCreateInstanceOperator(project_id=PROJECT_ID, body=CREATE_BODY, instance="", task_id="id")

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_create_should_validate_list_type(self, mock_hook):
        wrong_list_type_body = {
            "name": INSTANCE_NAME,
            "settings": {
                "tier": "db-n1-standard-1",
                "ipConfiguration": {
                    "authorizedNetworks": {}  # Should be a list, not a dict.
                    # Testing if the validation catches this.
                },
            },
        }
        op = CloudSQLCreateInstanceOperator(
            project_id=PROJECT_ID, body=wrong_list_type_body, instance=INSTANCE_NAME, task_id="id"
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert (
            "The field 'settings.ipConfiguration.authorizedNetworks' "
            "should be of list type according to the specification" in str(err)
        )
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_create_should_validate_non_empty_fields(self, mock_hook):
        empty_tier_body = {
            "name": INSTANCE_NAME,
            "settings": {
                "tier": "",  # Field can't be empty (defined in CLOUD_SQL_VALIDATION).
                # Testing if the validation catches this.
            },
        }
        op = CloudSQLCreateInstanceOperator(
            project_id=PROJECT_ID, body=empty_tier_body, instance=INSTANCE_NAME, task_id="id"
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert "The body field 'settings.tier' can't be empty. Please provide a value." in str(err)
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_patch(self, mock_hook):
        mock_hook.return_value.patch_instance.return_value = True
        op = CloudSQLInstancePatchOperator(
            project_id=PROJECT_ID, body=PATCH_BODY, instance=INSTANCE_NAME, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_instance.assert_called_once_with(
            project_id=PROJECT_ID, body=PATCH_BODY, instance=INSTANCE_NAME
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_patch_missing_project_id(self, mock_hook):
        mock_hook.return_value.patch_instance.return_value = True
        op = CloudSQLInstancePatchOperator(body=PATCH_BODY, instance=INSTANCE_NAME, task_id="id")
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_instance.assert_called_once_with(
            project_id=None, body=PATCH_BODY, instance=INSTANCE_NAME
        )
        assert result

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLInstancePatchOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_patch_should_bubble_up_ex_if_not_exists(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = False
        op = CloudSQLInstancePatchOperator(
            project_id=PROJECT_ID, body=PATCH_BODY, instance=INSTANCE_NAME, task_id="id"
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert "specify another instance to patch" in str(err)
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_instance.assert_not_called()

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_delete(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = True
        op = CloudSQLDeleteInstanceOperator(project_id=PROJECT_ID, instance=INSTANCE_NAME, task_id="id")
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql.CloudSQLCloneInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_clone(self, mock_hook, _check_if_instance_exists):
        destination_instance_name = "clone-test-name"
        _check_if_instance_exists.return_value = True
        op = CloudSQLCloneInstanceOperator(
            project_id=PROJECT_ID,
            instance=INSTANCE_NAME,
            destination_instance_name=destination_instance_name,
            task_id="id",
        )
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        body = {
            "cloneContext": {
                "kind": "sql#cloneContext",
                "destinationInstanceName": destination_instance_name,
            }
        }
        mock_hook.return_value.clone_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=body
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_delete_missing_project_id(self, mock_hook, _check_if_instance_exists):
        _check_if_instance_exists.return_value = True
        op = CloudSQLDeleteInstanceOperator(instance=INSTANCE_NAME, task_id="id")
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceOperator._check_if_instance_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_delete_should_abort_and_succeed_if_not_exists(
        self, mock_hook, _check_if_instance_exists
    ):
        _check_if_instance_exists.return_value = False
        op = CloudSQLDeleteInstanceOperator(project_id=PROJECT_ID, instance=INSTANCE_NAME, task_id="id")
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_not_called()

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_create(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = False
        op = CloudSQLCreateInstanceDatabaseOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=DATABASE_INSERT_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=DATABASE_INSERT_BODY
        )
        assert result

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_create_missing_project_id(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = False
        op = CloudSQLCreateInstanceDatabaseOperator(
            instance=INSTANCE_NAME, body=DATABASE_INSERT_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME, body=DATABASE_INSERT_BODY
        )
        assert result

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLCreateInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_create_should_abort_and_succeed_if_exists(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = True
        op = CloudSQLCreateInstanceDatabaseOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=DATABASE_INSERT_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_not_called()

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLPatchInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_patch(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = True
        op = CloudSQLPatchInstanceDatabaseOperator(
            project_id=PROJECT_ID,
            instance=INSTANCE_NAME,
            database=DB_NAME,
            body=DATABASE_PATCH_BODY,
            task_id="id",
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_database.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, database=DB_NAME, body=DATABASE_PATCH_BODY
        )
        assert result

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLPatchInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_patch_missing_project_id(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = True
        op = CloudSQLPatchInstanceDatabaseOperator(
            instance=INSTANCE_NAME, database=DB_NAME, body=DATABASE_PATCH_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_database.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME, database=DB_NAME, body=DATABASE_PATCH_BODY
        )
        assert result

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLPatchInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_patch_should_throw_ex_if_not_exists(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = False
        op = CloudSQLPatchInstanceDatabaseOperator(
            project_id=PROJECT_ID,
            instance=INSTANCE_NAME,
            database=DB_NAME,
            body=DATABASE_PATCH_BODY,
            task_id="id",
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert "Cloud SQL instance with ID" in str(err)
        assert "does not contain database" in str(err)
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.patch_database.assert_not_called()

    def test_instance_db_patch_should_throw_ex_when_empty_database(self):
        with pytest.raises(AirflowException, match="The required parameter 'database' is empty"):
            CloudSQLPatchInstanceDatabaseOperator(
                project_id=PROJECT_ID,
                instance=INSTANCE_NAME,
                database="",
                body=DATABASE_INSERT_BODY,
                task_id="id",
            )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_delete(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = True
        op = CloudSQLDeleteInstanceDatabaseOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, database=DB_NAME, task_id="id"
        )
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, database=DB_NAME
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_delete_missing_project_id(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = True
        op = CloudSQLDeleteInstanceDatabaseOperator(instance=INSTANCE_NAME, database=DB_NAME, task_id="id")
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME, database=DB_NAME
        )

    @mock.patch(
        "airflow.providers.google.cloud.operators.cloud_sql"
        ".CloudSQLDeleteInstanceDatabaseOperator._check_if_db_exists"
    )
    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_db_delete_should_abort_and_succeed_if_not_exists(self, mock_hook, _check_if_db_exists):
        _check_if_db_exists.return_value = False
        op = CloudSQLDeleteInstanceDatabaseOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, database=DB_NAME, task_id="id"
        )
        result = op.execute(None)
        assert result
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_export(self, mock_hook):
        mock_hook.return_value.export_instance.return_value = True
        op = CloudSQLExportInstanceOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=EXPORT_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.export_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=EXPORT_BODY
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_export_missing_project_id(self, mock_hook):
        mock_hook.return_value.export_instance.return_value = True
        op = CloudSQLExportInstanceOperator(instance=INSTANCE_NAME, body=EXPORT_BODY, task_id="id")
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.export_instance.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME, body=EXPORT_BODY
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    @mock.patch("airflow.providers.google.cloud.triggers.cloud_sql.CloudSQLAsyncHook")
    def test_execute_call_defer_method(self, mock_trigger_hook, mock_hook):
        operator = CloudSQLExportInstanceOperator(
            task_id="test_task",
            instance=INSTANCE_NAME,
            body=EXPORT_BODY,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as exc:
            operator.execute(mock.MagicMock())

        mock_hook.return_value.export_instance.assert_called_once()

        mock_hook.return_value.get_operation.assert_not_called()
        assert isinstance(exc.value.trigger, CloudSQLExportTrigger)
        assert exc.value.method_name == GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME

    def test_async_execute_should_should_throw_exception(self):
        """Tests that an AirflowException is raised in case of error event"""

        op = CloudSQLExportInstanceOperator(
            task_id="test_task",
            instance=INSTANCE_NAME,
            body=EXPORT_BODY,
            deferrable=True,
        )
        with pytest.raises(AirflowException):
            op.execute_complete(
                context=mock.MagicMock(), event={"status": "error", "message": "test failure message"}
            )

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_import(self, mock_hook):
        mock_hook.return_value.export_instance.return_value = True
        op = CloudSQLImportInstanceOperator(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=IMPORT_BODY, task_id="id"
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.import_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance=INSTANCE_NAME, body=IMPORT_BODY
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.cloud_sql.CloudSQLHook")
    def test_instance_import_missing_project_id(self, mock_hook):
        mock_hook.return_value.export_instance.return_value = True
        op = CloudSQLImportInstanceOperator(instance=INSTANCE_NAME, body=IMPORT_BODY, task_id="id")
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="v1beta4",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.import_instance.assert_called_once_with(
            project_id=None, instance=INSTANCE_NAME, body=IMPORT_BODY
        )
        assert result


class TestCloudSqlQueryValidation:
    @staticmethod
    def _setup_connections(get_connection, uri):
        gcp_connection = mock.MagicMock()
        gcp_connection.extra_dejson = mock.MagicMock()
        gcp_connection.extra_dejson.get.return_value = "empty_project"
        cloudsql_connection = Connection(uri=uri)
        cloudsql_connection2 = Connection(uri=uri)
        get_connection.side_effect = [gcp_connection, cloudsql_connection, cloudsql_connection2]

    @pytest.mark.parametrize(
        (
            "project_id",
            "location",
            "instance_name",
            "database_type",
            "use_proxy",
            "use_ssl",
            "sql",
            "message",
        ),
        [
            (
                "project_id",
                "",
                "instance_name",
                "mysql",
                False,
                False,
                "SELECT * FROM TEST",
                "The required extra 'location' is empty",
            ),
            (
                "project_id",
                "location",
                "",
                "postgres",
                False,
                False,
                "SELECT * FROM TEST",
                "The required extra 'instance' is empty",
            ),
            (
                "project_id",
                "location",
                "instance_name",
                "wrong",
                False,
                False,
                "SELECT * FROM TEST",
                "Invalid database type 'wrong'. Must be one of ['postgres', 'mysql']",
            ),
            (
                "project_id",
                "location",
                "instance_name",
                "postgres",
                True,
                True,
                "SELECT * FROM TEST",
                "Cloud SQL Proxy does not support SSL connections. SSL is not needed as"
                " Cloud SQL Proxy provides encryption on its own",
            ),
            (
                "project_id",
                "location",
                "instance_name",
                "postgres",
                False,
                True,
                "SELECT * FROM TEST",
                "SSL connections requires sslcert to be set",
            ),
        ],
    )
    @mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection")
    def test_create_operator_with_wrong_parameters(
        self,
        get_connection,
        project_id,
        location,
        instance_name,
        database_type,
        use_proxy,
        use_ssl,
        sql,
        message,
    ):
        uri = (
            f"gcpcloudsql://user:password@127.0.0.1:3200/testdb?"
            f"database_type={database_type}&project_id={project_id}&location={location}"
            f"&instance={instance_name}&use_proxy={use_proxy}&use_ssl={use_ssl}"
        )
        self._setup_connections(get_connection, uri)
        op = CloudSQLExecuteQueryOperator(sql=sql, task_id="task_id")
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert message in str(err)

    @mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection")
    def test_create_operator_with_too_long_unix_socket_path(self, get_connection):
        uri = (
            "gcpcloudsql://user:password@127.0.0.1:3200/testdb?database_type=postgres&"
            "project_id=example-project&location=europe-west1&"
            "instance=test_db_with_long_name_a_bit_above_the_limit_of_UNIX_socket_asdadadasadasd&"
            "use_proxy=True&sql_proxy_use_tcp=False"
        )
        self._setup_connections(get_connection, uri)
        operator = CloudSQLExecuteQueryOperator(sql=["SELECT * FROM TABLE"], task_id="task_id")
        with pytest.raises(AirflowException) as ctx:
            operator.execute(None)
        err = ctx.value
        assert "The UNIX socket path length cannot exceed" in str(err)

    @pytest.mark.parametrize(
        ("connection_port", "default_port", "expected_port"),
        [(None, 4321, 4321), (1234, None, 1234), (1234, 4321, 1234)],
    )
    def test_execute_openlineage_events(self, connection_port, default_port, expected_port):
        class DBApiHookForTests(DbApiHook):
            conn_name_attr = "sql_default"
            get_conn = MagicMock(name="conn")
            get_connection = MagicMock()

            def get_openlineage_database_info(self, connection):
                from airflow.providers.openlineage.sqlparser import DatabaseInfo

                return DatabaseInfo(
                    scheme="sqlscheme",
                    authority=DbApiHook.get_openlineage_authority_part(connection, default_port=default_port),
                )

        dbapi_hook = DBApiHookForTests()

        class CloudSQLExecuteQueryOperatorForTest(CloudSQLExecuteQueryOperator):
            @property
            def hook(self):
                return MagicMock(db_hook=dbapi_hook, database="")

        sql = """CREATE TABLE IF NOT EXISTS popular_orders_day_of_week (
            order_day_of_week VARCHAR(64) NOT NULL,
            order_placed_on   TIMESTAMP NOT NULL,
            orders_placed     INTEGER NOT NULL
        );
FORGOT TO COMMENT"""
        op = CloudSQLExecuteQueryOperatorForTest(task_id="task_id", sql=sql)
        DB_SCHEMA_NAME = "PUBLIC"
        rows = [
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "order_day_of_week", 1, "varchar"),
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "order_placed_on", 2, "timestamp"),
            (DB_SCHEMA_NAME, "popular_orders_day_of_week", "orders_placed", 3, "int4"),
        ]
        dbapi_hook.get_connection.return_value = Connection(
            conn_id="sql_default", conn_type="postgresql", host="host", port=connection_port
        )
        dbapi_hook.get_conn.return_value.cursor.return_value.fetchall.side_effect = [rows, []]

        lineage = op.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 0
        assert lineage.job_facets == {"sql": SQLJobFacet(query=sql)}
        assert lineage.run_facets["extractionError"].failedTasks == 1
        assert lineage.outputs == [
            Dataset(
                namespace=f"sqlscheme://host:{expected_port}",
                name="PUBLIC.popular_orders_day_of_week",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields(name="order_day_of_week", type="varchar"),
                            SchemaDatasetFacetFields(name="order_placed_on", type="timestamp"),
                            SchemaDatasetFacetFields(name="orders_placed", type="int4"),
                        ]
                    )
                },
            )
        ]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime, timedelta, timezone
from pathlib import Path
from unittest import mock

import pytest

from airflow.providers.common.compat.openlineage.facet import (
    Dataset,
    LifecycleStateChange,
    LifecycleStateChangeDatasetFacet,
    PreviousIdentifier,
)
from airflow.providers.google.cloud.operators.gcs import (
    GCSBucketCreateAclEntryOperator,
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSDeleteObjectsOperator,
    GCSFileTransformOperator,
    GCSListObjectsOperator,
    GCSObjectCreateAclEntryOperator,
    GCSSynchronizeBucketsOperator,
    GCSTimeSpanFileTransformOperator,
)

TASK_ID = "test-gcs-operator"
TEST_BUCKET = "test-bucket"
TEST_PROJECT = "test-project"
DELIMITER = ".csv"
PREFIX = "TEST"
PREFIX_2 = "TEST2"
MOCK_FILES = ["TEST1.csv", "TEST2.csv", "TEST3.csv", "OTHERTEST1.csv"]
TEST_OBJECT = "dir1/test-object"
LOCAL_FILE_PATH = "/home/airflow/gcp/test-object"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class TestGoogleCloudStorageCreateBucket:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute(self, mock_hook):
        operator = GCSCreateBucketOperator(
            task_id=TASK_ID,
            bucket_name=TEST_BUCKET,
            resource={"lifecycle": {"rule": [{"action": {"type": "Delete"}, "condition": {"age": 7}}]}},
            storage_class="MULTI_REGIONAL",
            location="EU",
            labels={"env": "prod"},
            project_id=TEST_PROJECT,
        )

        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.create_bucket.assert_called_once_with(
            bucket_name=TEST_BUCKET,
            storage_class="MULTI_REGIONAL",
            location="EU",
            labels={"env": "prod"},
            project_id=TEST_PROJECT,
            resource={"lifecycle": {"rule": [{"action": {"type": "Delete"}, "condition": {"age": 7}}]}},
        )


class TestGoogleCloudStorageAcl:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_bucket_create_acl(self, mock_hook):
        operator = GCSBucketCreateAclEntryOperator(
            bucket="test-bucket",
            entity="test-entity",
            role="test-role",
            user_project="test-user-project",
            task_id="id",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.insert_bucket_acl.assert_called_once_with(
            bucket_name="test-bucket",
            entity="test-entity",
            role="test-role",
            user_project="test-user-project",
        )

    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_object_create_acl(self, mock_hook):
        operator = GCSObjectCreateAclEntryOperator(
            bucket="test-bucket",
            object_name="test-object",
            entity="test-entity",
            generation=42,
            role="test-role",
            user_project="test-user-project",
            task_id="id",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.insert_object_acl.assert_called_once_with(
            bucket_name="test-bucket",
            object_name="test-object",
            entity="test-entity",
            generation=42,
            role="test-role",
            user_project="test-user-project",
        )


class TestGCSDeleteObjectsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_delete_objects(self, mock_hook):
        operator = GCSDeleteObjectsOperator(task_id=TASK_ID, bucket_name=TEST_BUCKET, objects=MOCK_FILES[0:2])

        operator.execute(None)
        mock_hook.return_value.list.assert_not_called()
        mock_hook.return_value.delete.assert_has_calls(
            calls=[
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[0]),
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[1]),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_delete_empty_list_of_objects(self, mock_hook):
        operator = GCSDeleteObjectsOperator(task_id=TASK_ID, bucket_name=TEST_BUCKET, objects=[])

        operator.execute(None)
        mock_hook.return_value.list.assert_not_called()
        mock_hook.return_value.delete.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_delete_prefix(self, mock_hook):
        mock_hook.return_value.list.return_value = MOCK_FILES[1:4]
        operator = GCSDeleteObjectsOperator(task_id=TASK_ID, bucket_name=TEST_BUCKET, prefix=PREFIX)

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(bucket_name=TEST_BUCKET, prefix=PREFIX)
        mock_hook.return_value.delete.assert_has_calls(
            calls=[
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[1]),
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[2]),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_delete_prefix_as_empty_string(self, mock_hook):
        mock_hook.return_value.list.return_value = MOCK_FILES[0:4]
        operator = GCSDeleteObjectsOperator(task_id=TASK_ID, bucket_name=TEST_BUCKET, prefix="")

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(bucket_name=TEST_BUCKET, prefix="")
        mock_hook.return_value.delete.assert_has_calls(
            calls=[
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[0]),
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[1]),
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[2]),
                mock.call(bucket_name=TEST_BUCKET, object_name=MOCK_FILES[3]),
            ],
            any_order=True,
        )

    @pytest.mark.parametrize(
        ("objects", "prefix", "inputs"),
        (
            (["folder/a.txt", "b.json"], None, ["folder/a.txt", "b.json"]),
            (["folder/a.txt", "folder/b.json"], None, ["folder/a.txt", "folder/b.json"]),
            (None, ["folder/a.txt", "b.json"], ["folder/a.txt", "b.json"]),
            (None, "dir/pre", ["dir"]),
            (None, ["dir/"], ["dir"]),
            (None, "", ["/"]),
            (None, "/", ["/"]),
            (None, "pre", ["/"]),
            (None, "dir/pre*", ["dir"]),
            (None, "*", ["/"]),
        ),
        ids=(
            "objects",
            "multiple objects in the same dir",
            "objects as prefixes",
            "directory with prefix",
            "directory",
            "empty prefix",
            "slash as prefix",
            "prefix with no ending slash",
            "directory with prefix with wildcard",
            "just wildcard",
        ),
    )
    def test_get_openlineage_facets_on_start(self, objects, prefix, inputs):
        bucket_url = f"gs://{TEST_BUCKET}"
        expected_inputs = [
            Dataset(
                namespace=bucket_url,
                name=name,
                facets={
                    "lifecycleStateChange": LifecycleStateChangeDatasetFacet(
                        lifecycleStateChange=LifecycleStateChange.DROP.value,
                        previousIdentifier=PreviousIdentifier(
                            namespace=bucket_url,
                            name=name,
                        ),
                    )
                },
            )
            for name in inputs
        ]

        operator = GCSDeleteObjectsOperator(
            task_id=TASK_ID, bucket_name=TEST_BUCKET, objects=objects, prefix=prefix
        )
        lineage = operator.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == len(inputs)
        assert len(lineage.outputs) == 0
        assert all(element in lineage.inputs for element in expected_inputs)
        assert all(element in expected_inputs for element in lineage.inputs)


class TestGoogleCloudStorageListOperator:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute__delimiter(self, mock_hook):
        mock_hook.return_value.list.return_value = MOCK_FILES
        operator = GCSListObjectsOperator(
            task_id=TASK_ID, bucket=TEST_BUCKET, prefix=PREFIX, delimiter=DELIMITER
        )
        files = operator.execute(context=mock.MagicMock())
        mock_hook.return_value.list.assert_called_once_with(
            bucket_name=TEST_BUCKET, prefix=PREFIX, delimiter=DELIMITER, match_glob=None
        )
        assert sorted(files) == sorted(MOCK_FILES)

    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute__match_glob(self, mock_hook):
        mock_hook.return_value.list.return_value = MOCK_FILES
        operator = GCSListObjectsOperator(
            task_id=TASK_ID, bucket=TEST_BUCKET, prefix=PREFIX, match_glob=f"**/*{DELIMITER}", delimiter=None
        )
        files = operator.execute(context=mock.MagicMock())
        mock_hook.return_value.list.assert_called_once_with(
            bucket_name=TEST_BUCKET, prefix=PREFIX, match_glob=f"**/*{DELIMITER}", delimiter=None
        )
        assert sorted(files) == sorted(MOCK_FILES)


class TestGCSFileTransformOperator:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.NamedTemporaryFile")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.subprocess")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute(self, mock_hook, mock_subprocess, mock_tempfile):
        source_bucket = TEST_BUCKET
        source_object = "test.txt"
        destination_bucket = TEST_BUCKET + "-dest"
        destination_object = "transformed_test.txt"
        transform_script = "script.py"

        source = "source"
        destination = "destination"

        # Mock the name attribute...
        mock1 = mock.Mock()
        mock2 = mock.Mock()
        mock1.name = source
        mock2.name = destination

        mock_tempfile.return_value.__enter__.side_effect = [mock1, mock2]

        mock_proc = mock.MagicMock()
        mock_proc.returncode = 0
        mock_proc.stdout.readline = lambda: b""
        mock_proc.wait.return_value = None
        mock_popen = mock.MagicMock()
        mock_popen.return_value.__enter__.return_value = mock_proc

        mock_subprocess.Popen = mock_popen
        mock_subprocess.PIPE = "pipe"
        mock_subprocess.STDOUT = "stdout"

        op = GCSFileTransformOperator(
            task_id=TASK_ID,
            source_bucket=source_bucket,
            source_object=source_object,
            destination_object=destination_object,
            destination_bucket=destination_bucket,
            transform_script=transform_script,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.download.assert_called_once_with(
            bucket_name=source_bucket, object_name=source_object, filename=source
        )

        mock_subprocess.Popen.assert_called_once_with(
            args=[transform_script, source, destination],
            stdout="pipe",
            stderr="stdout",
            close_fds=True,
        )

        mock_hook.return_value.upload.assert_called_with(
            bucket_name=destination_bucket,
            object_name=destination_object,
            filename=destination,
        )

    def test_get_openlineage_facets_on_start(self):
        expected_input = Dataset(
            namespace=f"gs://{TEST_BUCKET}",
            name="folder/a.txt",
        )
        expected_output = Dataset(
            namespace=f"gs://{TEST_BUCKET}2",
            name="b.txt",
        )

        operator = GCSFileTransformOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object="folder/a.txt",
            destination_bucket=f"{TEST_BUCKET}2",
            destination_object="b.txt",
            transform_script="/path/to_script",
        )

        lineage = operator.get_openlineage_facets_on_start()
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == expected_input
        assert lineage.outputs[0] == expected_output


class TestGCSTimeSpanFileTransformOperatorDateInterpolation:
    def test_execute(self):
        interp_dt = datetime(2015, 2, 1, 15, 16, 17, 345, tzinfo=timezone.utc)

        assert GCSTimeSpanFileTransformOperator.interpolate_prefix(None, interp_dt) is None

        assert (
            GCSTimeSpanFileTransformOperator.interpolate_prefix("prefix_without_date", interp_dt)
            == "prefix_without_date"
        )

        assert (
            GCSTimeSpanFileTransformOperator.interpolate_prefix("prefix_with_year_%Y", interp_dt)
            == "prefix_with_year_2015"
        )

        assert (
            GCSTimeSpanFileTransformOperator.interpolate_prefix(
                "prefix_with_year_month_day/%Y/%m/%d/", interp_dt
            )
            == "prefix_with_year_month_day/2015/02/01/"
        )

        assert (
            GCSTimeSpanFileTransformOperator.interpolate_prefix(
                "prefix_with_year_month_day_and_percent_%%/%Y/%m/%d/", interp_dt
            )
            == "prefix_with_year_month_day_and_percent_%/2015/02/01/"
        )


class TestGCSTimeSpanFileTransformOperator:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.TemporaryDirectory")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.subprocess")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute(self, mock_hook, mock_subprocess, mock_tempdir):
        source_bucket = TEST_BUCKET
        source_prefix = "source_prefix"
        source_gcp_conn_id = ""

        destination_bucket = TEST_BUCKET + "_dest"
        destination_prefix = "destination_prefix"
        destination_gcp_conn_id = ""

        transform_script = "script.py"

        source = "source"
        destination = "destination"

        file1 = "file1"
        file2 = "file2"

        timespan_start = datetime(2015, 2, 1, 15, 16, 17, 345, tzinfo=timezone.utc)
        timespan_end = timespan_start + timedelta(hours=1)

        mock_ti = mock.Mock()
        context = dict(
            logical_date=timespan_start,
            data_interval_start=timespan_start,
            data_interval_end=timespan_end,
            ti=mock_ti,
            task=mock.MagicMock(),
        )

        mock_tempdir.return_value.__enter__.side_effect = [source, destination]
        mock_hook.return_value.list_by_timespan.return_value = [
            f"{source_prefix}/{file1}",
            f"{source_prefix}/{file2}",
        ]

        mock_proc = mock.MagicMock()
        mock_proc.returncode = 0
        mock_proc.stdout.readline = lambda: b""
        mock_proc.wait.return_value = None
        mock_popen = mock.MagicMock()
        mock_popen.return_value.__enter__.return_value = mock_proc

        mock_subprocess.Popen = mock_popen
        mock_subprocess.PIPE = "pipe"
        mock_subprocess.STDOUT = "stdout"

        op = GCSTimeSpanFileTransformOperator(
            task_id=TASK_ID,
            source_bucket=source_bucket,
            source_prefix=source_prefix,
            source_gcp_conn_id=source_gcp_conn_id,
            destination_bucket=destination_bucket,
            destination_prefix=destination_prefix,
            destination_gcp_conn_id=destination_gcp_conn_id,
            transform_script=transform_script,
        )

        with mock.patch.object(Path, "glob") as path_glob:
            path_glob.return_value.__iter__.return_value = [
                Path(f"{destination}/{file1}"),
                Path(f"{destination}/{file2}"),
            ]
            op.execute(context=context)

        mock_hook.return_value.list_by_timespan.assert_called_once_with(
            bucket_name=source_bucket,
            timespan_start=timespan_start,
            timespan_end=timespan_end,
            prefix=source_prefix,
        )

        mock_hook.return_value.download.assert_has_calls(
            [
                mock.call(
                    bucket_name=source_bucket,
                    object_name=f"{source_prefix}/{file1}",
                    filename=f"{source}/{source_prefix}/{file1}",
                    chunk_size=None,
                    num_max_attempts=1,
                ),
                mock.call(
                    bucket_name=source_bucket,
                    object_name=f"{source_prefix}/{file2}",
                    filename=f"{source}/{source_prefix}/{file2}",
                    chunk_size=None,
                    num_max_attempts=1,
                ),
            ]
        )

        mock_subprocess.Popen.assert_called_once_with(
            args=[
                transform_script,
                source,
                destination,
                timespan_start.replace(microsecond=0).isoformat(),
                timespan_end.replace(microsecond=0).isoformat(),
            ],
            stdout="pipe",
            stderr="stdout",
            close_fds=True,
        )

        mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(
                    bucket_name=destination_bucket,
                    filename=f"{destination}/{file1}",
                    object_name=f"{destination_prefix}/{file1}",
                    chunk_size=None,
                    num_max_attempts=1,
                ),
                mock.call(
                    bucket_name=destination_bucket,
                    filename=f"{destination}/{file2}",
                    object_name=f"{destination_prefix}/{file2}",
                    chunk_size=None,
                    num_max_attempts=1,
                ),
            ]
        )

    @pytest.mark.parametrize(
        ("source_prefix", "dest_prefix", "inputs", "outputs"),
        (
            (
                None,
                None,
                [Dataset(f"gs://{TEST_BUCKET}", "/")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "/")],
            ),
            (
                None,
                "dest_pre/",
                [Dataset(f"gs://{TEST_BUCKET}", "/")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "dest_pre")],
            ),
            (
                "source_pre/",
                None,
                [Dataset(f"gs://{TEST_BUCKET}", "source_pre")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "/")],
            ),
            (
                "source_pre/",
                "dest_pre/",
                [Dataset(f"gs://{TEST_BUCKET}", "source_pre")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "dest_pre")],
            ),
            (
                "source_pre",
                "dest_pre",
                [Dataset(f"gs://{TEST_BUCKET}", "/")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "/")],
            ),
            (
                "dir1/source_pre",
                "dir2/dest_pre",
                [Dataset(f"gs://{TEST_BUCKET}", "dir1")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "dir2")],
            ),
            (
                "",
                "/",
                [Dataset(f"gs://{TEST_BUCKET}", "/")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "/")],
            ),
            (
                "source/a.txt",
                "target/",
                [Dataset(f"gs://{TEST_BUCKET}", "source/a.txt")],
                [Dataset(f"gs://{TEST_BUCKET}_dest", "target")],
            ),
        ),
        ids=(
            "no prefixes",
            "dest prefix only",
            "source prefix only",
            "both with ending slash",
            "both without ending slash",
            "both as directory with prefix",
            "both empty or root",
            "source prefix is file path",
        ),
    )
    @mock.patch("airflow.providers.google.cloud.operators.gcs.TemporaryDirectory")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.subprocess")
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_get_openlineage_facets_on_complete(
        self, mock_hook, mock_subprocess, mock_tempdir, source_prefix, dest_prefix, inputs, outputs
    ):
        source_bucket = TEST_BUCKET

        destination_bucket = TEST_BUCKET + "_dest"
        destination = "destination"

        file1 = "file1"
        file2 = "file2"

        timespan_start = datetime(2015, 2, 1, 15, 16, 17, 345, tzinfo=timezone.utc)
        timespan_end = timespan_start + timedelta(hours=1)

        context = dict(
            logical_date=timespan_start,
            data_interval_start=timespan_start,
            data_interval_end=timespan_end,
            ti=mock.Mock(),
            task=mock.MagicMock(),
        )

        mock_tempdir.return_value.__enter__.side_effect = ["source", destination]
        mock_hook.return_value.list_by_timespan.return_value = [
            f"{source_prefix or ''}{file1}",
            f"{source_prefix or ''}{file2}",
        ]

        mock_proc = mock.MagicMock()
        mock_proc.returncode = 0
        mock_proc.stdout.readline = lambda: b""
        mock_proc.wait.return_value = None
        mock_popen = mock.MagicMock()
        mock_popen.return_value.__enter__.return_value = mock_proc

        mock_subprocess.Popen = mock_popen
        mock_subprocess.PIPE = "pipe"
        mock_subprocess.STDOUT = "stdout"

        op = GCSTimeSpanFileTransformOperator(
            task_id=TASK_ID,
            source_bucket=source_bucket,
            source_prefix=source_prefix,
            source_gcp_conn_id="",
            destination_bucket=destination_bucket,
            destination_prefix=dest_prefix,
            destination_gcp_conn_id="",
            transform_script="script.py",
        )

        with mock.patch.object(Path, "glob") as path_glob:
            path_glob.return_value.__iter__.return_value = [
                Path(f"{destination}/{file1}"),
                Path(f"{destination}/{file2}"),
            ]
            op.execute(context=context)

        lineage = op.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == len(inputs)
        assert len(lineage.outputs) == len(outputs)
        assert all(element in lineage.inputs for element in inputs)
        assert all(element in inputs for element in lineage.inputs)
        assert all(element in lineage.outputs for element in outputs)
        assert all(element in outputs for element in lineage.outputs)


class TestGCSDeleteBucketOperator:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_delete_bucket(self, mock_hook):
        operator = GCSDeleteBucketOperator(task_id=TASK_ID, bucket_name=TEST_BUCKET)

        operator.execute(None)
        mock_hook.return_value.delete_bucket.assert_called_once_with(
            bucket_name=TEST_BUCKET, force=True, user_project=None
        )


class TestGoogleCloudStorageSync:
    @mock.patch("airflow.providers.google.cloud.operators.gcs.GCSHook")
    def test_execute(self, mock_hook):
        task = GCSSynchronizeBucketsOperator(
            task_id="task-id",
            source_bucket="SOURCE_BUCKET",
            destination_bucket="DESTINATION_BUCKET",
            source_object="SOURCE_OBJECT",
            destination_object="DESTINATION_OBJECT",
            recursive=True,
            delete_extra_files=True,
            allow_overwrite=True,
            gcp_conn_id="GCP_CONN_ID",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        task.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id="GCP_CONN_ID",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.sync.assert_called_once_with(
            source_bucket="SOURCE_BUCKET",
            source_object="SOURCE_OBJECT",
            destination_bucket="DESTINATION_BUCKET",
            destination_object="DESTINATION_OBJECT",
            delete_extra_files=True,
            recursive=True,
            allow_overwrite=True,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.google.common.links.storage import FileDetailsLink, StorageLink


class TestStorageLink:
    def test_storage_link(self):
        assert StorageLink.name == "GCS Storage"
        assert StorageLink.key == "storage_conf"
        assert (
            StorageLink.format_str
            == "https://console.cloud.google.com/storage/browser/{uri};tab=objects?project={project_id}"
        )

    def test_storage_link_format(self):
        link = StorageLink()
        url = link._format_link(uri="test-bucket/test-folder", project_id="test-id")
        expected_url = "https://console.cloud.google.com/storage/browser/test-bucket/test-folder;tab=objects?project=test-id"
        assert url == expected_url


class TestFileDetailsLink:
    def test_file_details_link_name_and_key(self):
        assert FileDetailsLink.name == "GCS File Details"
        assert FileDetailsLink.key == "file_details"
        assert (
            FileDetailsLink.format_str
            == "https://console.cloud.google.com/storage/browser/_details/{uri};tab=live_object?project={project_id}"
        )

    def test_file_details_link_format(self):
        link = FileDetailsLink()
        url = link._format_link(uri="test-bucket/test-folder", project_id="test-id")
        expected_url = "https://console.cloud.google.com/storage/browser/_details/test-bucket/test-folder;tab=live_object?project=test-id"
        assert url == expected_url



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains various unit tests for
functions in CloudBuildHook
"""

from __future__ import annotations

from concurrent.futures import Future
from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.devtools.cloudbuild_v1 import CloudBuildAsyncClient, GetBuildRequest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.cloud_build import CloudBuildAsyncHook, CloudBuildHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_no_default_project_id

PROJECT_ID = "cloud-build-project"
LOCATION = "test-location"
PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}"
CLOUD_BUILD_PATH = "airflow.providers.google.cloud.hooks.cloud_build.{}"
BUILD_ID = "test-build-id-9832662"
REPO_SOURCE = {"repo_source": {"repo_name": "test_repo", "branch_name": "main"}}
BUILD = {
    "source": REPO_SOURCE,
    "steps": [{"name": "gcr.io/cloud-builders/gcloud", "entrypoint": "/bin/sh", "args": ["-c", "ls"]}],
    "status": "SUCCESS",
}
BUILD_WORKING = {
    "source": REPO_SOURCE,
    "steps": [{"name": "gcr.io/cloud-builders/gcloud", "entrypoint": "/bin/sh", "args": ["-c", "ls"]}],
    "status": "WORKING",
}
BUILD_TRIGGER = {
    "name": "test-cloud-build-trigger",
    "trigger_template": {"project_id": PROJECT_ID, "repo_name": "test_repo", "branch_name": "main"},
    "filename": "cloudbuild.yaml",
}
OPERATION = {"metadata": {"build": {"id": BUILD_ID}}}
TRIGGER_ID = "32488e7f-09d6-4fe9-a5fb-4ca1419a6e7a"


class TestCloudBuildHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.hook = CloudBuildHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildClient")
    def test_cloud_build_service_client_creation(self, mock_client, mock_get_creds):
        result = self.hook.get_conn()
        mock_client.assert_called_once_with(
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            client_options=None,
        )
        assert mock_client.return_value == result
        assert self.hook._client["global"] == result

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_cancel_build(self, get_conn):
        self.hook.cancel_build(id_=BUILD_ID, project_id=PROJECT_ID)

        get_conn.return_value.cancel_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook._get_build_id_from_operation"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_create_build_without_waiting_for_result(self, get_conn, mock_get_id_from_operation):
        get_conn.return_value.run_build_trigger.return_value = mock.MagicMock()
        mock_get_id_from_operation.return_value = BUILD_ID

        self.hook.create_build_without_waiting_for_result(
            build=BUILD, project_id=PROJECT_ID, location=LOCATION
        )

        mock_operation = get_conn.return_value.create_build

        mock_operation.assert_called_once_with(
            request={"parent": PARENT, "project_id": PROJECT_ID, "build": BUILD},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        mock_get_id_from_operation.assert_called_once_with(mock_operation())

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_create_build_trigger(self, get_conn):
        self.hook.create_build_trigger(trigger=BUILD_TRIGGER, project_id=PROJECT_ID)

        get_conn.return_value.create_build_trigger.assert_called_once_with(
            request={"project_id": PROJECT_ID, "trigger": BUILD_TRIGGER},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_delete_build_trigger(self, get_conn):
        self.hook.delete_build_trigger(trigger_id=TRIGGER_ID, project_id=PROJECT_ID)

        get_conn.return_value.delete_build_trigger.assert_called_once_with(
            request={"project_id": PROJECT_ID, "trigger_id": TRIGGER_ID},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_get_build(self, get_conn):
        self.hook.get_build(id_=BUILD_ID, project_id=PROJECT_ID)

        get_conn.return_value.get_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_get_build_trigger(self, get_conn):
        self.hook.get_build_trigger(trigger_id=TRIGGER_ID, project_id=PROJECT_ID)

        get_conn.return_value.get_build_trigger.assert_called_once_with(
            request={"project_id": PROJECT_ID, "trigger_id": TRIGGER_ID},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_list_build_triggers(self, get_conn):
        self.hook.list_build_triggers(project_id=PROJECT_ID, location=LOCATION)

        get_conn.return_value.list_build_triggers.assert_called_once_with(
            request={"parent": PARENT, "project_id": PROJECT_ID, "page_size": None, "page_token": None},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_list_builds(self, get_conn):
        self.hook.list_builds(project_id=PROJECT_ID, location=LOCATION)

        get_conn.return_value.list_builds.assert_called_once_with(
            request={
                "parent": PARENT,
                "project_id": PROJECT_ID,
                "page_size": None,
                "page_token": None,
                "filter": None,
            },
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook._get_build_id_from_operation"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.TIME_TO_SLEEP_IN_SECONDS")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_retry_build_with_wait(self, get_conn, wait_time, mock_get_id_from_operation):
        get_conn.return_value.run_build_trigger.return_value = mock.MagicMock()
        mock_get_id_from_operation.return_value = BUILD_ID

        wait_time.return_value = 0

        self.hook.retry_build(id_=BUILD_ID, project_id=PROJECT_ID)

        mock_operation = get_conn.return_value.retry_build

        mock_operation.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

        get_conn.return_value.retry_build.return_value.result.assert_called_once_with(
            timeout=None,
            polling=None,
            retry=None,
        )

        get_conn.return_value.get_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

        mock_get_id_from_operation.assert_called_once_with(mock_operation())

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook._get_build_id_from_operation"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_retry_build_without_wait(self, get_conn, mock_get_id_from_operation):
        get_conn.return_value.run_build_trigger.return_value = mock.MagicMock()
        mock_get_id_from_operation.return_value = BUILD_ID

        self.hook.retry_build(id_=BUILD_ID, project_id=PROJECT_ID, wait=False)

        get_conn.return_value.retry_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

        get_conn.return_value.get_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook._get_build_id_from_operation"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.TIME_TO_SLEEP_IN_SECONDS")
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_run_build_trigger_with_wait(self, get_conn, wait_time, mock_get_id_from_operation):
        get_conn.return_value.run_build_trigger.return_value = mock.MagicMock()
        mock_get_id_from_operation.return_value = BUILD_ID

        wait_time.return_value = 0

        self.hook.run_build_trigger(
            trigger_id=TRIGGER_ID, source=REPO_SOURCE["repo_source"], project_id=PROJECT_ID
        )

        mock_operation = get_conn.return_value.run_build_trigger

        mock_operation.assert_called_once_with(
            request={
                "project_id": PROJECT_ID,
                "trigger_id": TRIGGER_ID,
                "source": REPO_SOURCE["repo_source"],
            },
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        get_conn.return_value.run_build_trigger.return_value.result.assert_called_once_with(
            timeout=None,
            polling=None,
            retry=None,
        )

        get_conn.return_value.get_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

        mock_get_id_from_operation.assert_called_once_with(mock_operation())

    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook._get_build_id_from_operation"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_run_build_trigger_without_wait(self, get_conn, mock_get_id_from_operation):
        get_conn.return_value.run_build_trigger.return_value = mock.MagicMock()
        mock_get_id_from_operation.return_value = BUILD_ID

        self.hook.run_build_trigger(
            trigger_id=TRIGGER_ID, source=REPO_SOURCE["repo_source"], project_id=PROJECT_ID, wait=False
        )

        get_conn.return_value.run_build_trigger.assert_called_once_with(
            request={
                "project_id": PROJECT_ID,
                "trigger_id": TRIGGER_ID,
                "source": REPO_SOURCE["repo_source"],
            },
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        get_conn.return_value.get_build.assert_called_once_with(
            request={"project_id": PROJECT_ID, "id": BUILD_ID}, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook.get_conn")
    def test_update_build_trigger(self, get_conn):
        self.hook.update_build_trigger(trigger_id=TRIGGER_ID, trigger=BUILD_TRIGGER, project_id=PROJECT_ID)

        get_conn.return_value.update_build_trigger.assert_called_once_with(
            request={"project_id": PROJECT_ID, "trigger_id": TRIGGER_ID, "trigger": BUILD_TRIGGER},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


@pytest.mark.db_test
class TestAsyncHook:
    @pytest.fixture
    def hook(self):
        return CloudBuildAsyncHook(
            gcp_conn_id="google_cloud_default",
        )

    @pytest.mark.asyncio
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncHook.get_credentials"))
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncClient.get_build"))
    async def test_async_cloud_build_service_client_creation_should_execute_successfully(
        self, mocked_get_build, mock_get_creds, hook, mocker
    ):
        fake_credentials = mock.MagicMock(name="FakeCreds")
        mock_get_creds.return_value = fake_credentials
        mocked_cloud_build_async_client = mocker.patch.object(
            CloudBuildAsyncClient, "__init__", return_value=None, spec=CloudBuildAsyncClient
        )
        mocked_get_build.return_value = Future()

        await hook.get_cloud_build(project_id=PROJECT_ID, id_=BUILD_ID)
        request = GetBuildRequest(
            dict(
                project_id=PROJECT_ID,
                id=BUILD_ID,
            )
        )

        mock_get_creds.assert_called_once()
        mocked_get_build.assert_called_once_with(request=request, retry=DEFAULT, timeout=None, metadata=())
        mocked_cloud_build_async_client.assert_called_once_with(
            client_info=mock.ANY, client_options=None, credentials=fake_credentials
        )

    @pytest.mark.asyncio
    async def test_async_get_clod_build_without_build_id_should_throw_exception(self, hook):
        with pytest.raises(AirflowException, match=r"Google Cloud Build id is required."):
            await hook.get_cloud_build(project_id=PROJECT_ID, id_=None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

resources = pytest.importorskip("google.cloud.aiplatform.vertex_ray.util.resources")
from google.cloud.aiplatform.vertex_ray.util.resources import Cluster, Resources

from airflow.providers.google.cloud.hooks.vertex_ai.ray import RayHook

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_LOCATION: str = "test-location"
TEST_PROJECT_ID: str = "test-project-id"
TEST_NODE_RESOURCES: dict = {
    "machine_type": "n1-standard-8",
    "node_count": 1,
    "accelerator_type": "NVIDIA_TESLA_K80",
    "accelerator_count": 1,
    "custom_image": "us-docker.pkg.dev/my-project/ray-cpu-image.2.9:latest",
}
TEST_PYTHON_VERSION: str = "3.10"
TEST_RAY_VERSION: str = "2.33"
TEST_CLUSTER_NAME: str = "test-cluster-name"

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
RAY_STRING = "airflow.providers.google.cloud.hooks.vertex_ai.ray.{}"


class TestRayWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = RayHook(gcp_conn_id=TEST_GCP_CONN_ID)
            self.hook.get_credentials = mock.MagicMock()

    @mock.patch(RAY_STRING.format("vertex_ray.create_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    def test_create_ray_cluster(self, mock_aiplatform_init, mock_create_ray_cluster) -> None:
        self.hook.create_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            head_node_type=TEST_NODE_RESOURCES,
            python_version=TEST_PYTHON_VERSION,
            ray_version=TEST_RAY_VERSION,
            network=None,
            service_account=None,
            cluster_name=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
            custom_images=None,
            enable_metrics_collection=True,
            enable_logging=True,
            psc_interface_config=None,
            reserved_ip_ranges=None,
            labels=None,
        )
        mock_aiplatform_init.assert_called_once()
        mock_create_ray_cluster.assert_called_once_with(
            head_node_type=TEST_NODE_RESOURCES,
            python_version=TEST_PYTHON_VERSION,
            ray_version=TEST_RAY_VERSION,
            network=None,
            service_account=None,
            cluster_name=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
            custom_images=None,
            enable_metrics_collection=True,
            enable_logging=True,
            psc_interface_config=None,
            reserved_ip_ranges=None,
            labels=None,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.delete_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_delete_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_delete_ray_cluster
    ) -> None:
        self.hook.delete_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_delete_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.get_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_get_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_get_ray_cluster
    ) -> None:
        self.hook.get_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_get_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.update_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_update_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_update_ray_cluster
    ) -> None:
        self.hook.update_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_update_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
            worker_node_types=[TEST_NODE_RESOURCES],
        )

    @mock.patch(RAY_STRING.format("vertex_ray.list_ray_clusters"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    def test_list_ray_clusters(self, mock_aiplatform_init, mock_list_ray_clusters) -> None:
        self.hook.list_ray_clusters(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
        )
        mock_aiplatform_init.assert_called_once()
        mock_list_ray_clusters.assert_called_once()

    @mock.patch(RAY_STRING.format("aiplatform.init"))
    def test_serialize_cluster_obj(self, mock_aiplatform_init) -> None:
        RESOURCE_SAMPLE = {
            "accelerator_count": 0,
            "accelerator_type": None,
            "autoscaling_spec": None,
            "boot_disk_size_gb": 100,
            "boot_disk_type": "pd-ssd",
            "custom_image": None,
            "machine_type": "n1-standard-16",
            "node_count": 1,
        }
        SAMPLE_CLUSTER_SERIALIZED = {
            "cluster_resource_name": TEST_CLUSTER_NAME,
            "dashboard_address": "dashboard_addr",
            "head_node_type": RESOURCE_SAMPLE,
            "labels": {"label1": "val1"},
            "network": "custom_network",
            "psc_interface_config": None,
            "python_version": TEST_PYTHON_VERSION,
            "ray_logs_enabled": True,
            "ray_metric_enabled": True,
            "ray_version": TEST_RAY_VERSION,
            "reserved_ip_ranges": [
                "172.16.0.0/16",
                "10.10.10.0/28",
            ],
            "service_account": None,
            "state": "RUNNING",
            "worker_node_types": [RESOURCE_SAMPLE, RESOURCE_SAMPLE],
        }
        cluster_obj = Cluster(
            cluster_resource_name=TEST_CLUSTER_NAME,
            state="RUNNING",  # type: ignore[arg-type]
            network="custom_network",
            reserved_ip_ranges=["172.16.0.0/16", "10.10.10.0/28"],
            python_version=TEST_PYTHON_VERSION,
            ray_version=TEST_RAY_VERSION,
            head_node_type=Resources(**RESOURCE_SAMPLE),  # type: ignore[arg-type]
            worker_node_types=[
                Resources(**RESOURCE_SAMPLE),  # type: ignore[arg-type]
                Resources(**RESOURCE_SAMPLE),  # type: ignore[arg-type]
            ],
            dashboard_address="dashboard_addr",
            labels={"label1": "val1"},
        )

        assert self.hook.serialize_cluster_obj(cluster_obj) == SAMPLE_CLUSTER_SERIALIZED


class TestRayWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = RayHook(gcp_conn_id=TEST_GCP_CONN_ID)
            self.hook.get_credentials = mock.MagicMock()

    @mock.patch(RAY_STRING.format("vertex_ray.create_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    def test_create_ray_cluster(self, mock_aiplatform_init, mock_create_ray_cluster) -> None:
        self.hook.create_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            head_node_type=TEST_NODE_RESOURCES,
            python_version=TEST_PYTHON_VERSION,
            ray_version=TEST_RAY_VERSION,
            network=None,
            service_account=None,
            cluster_name=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
            custom_images=None,
            enable_metrics_collection=True,
            enable_logging=True,
            psc_interface_config=None,
            reserved_ip_ranges=None,
            labels=None,
        )
        mock_aiplatform_init.assert_called_once()
        mock_create_ray_cluster.assert_called_once_with(
            head_node_type=TEST_NODE_RESOURCES,
            python_version=TEST_PYTHON_VERSION,
            ray_version=TEST_RAY_VERSION,
            network=None,
            service_account=None,
            cluster_name=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
            custom_images=None,
            enable_metrics_collection=True,
            enable_logging=True,
            psc_interface_config=None,
            reserved_ip_ranges=None,
            labels=None,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.delete_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_delete_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_delete_ray_cluster
    ) -> None:
        self.hook.delete_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_delete_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.get_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_get_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_get_ray_cluster
    ) -> None:
        self.hook.get_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_get_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
        )

    @mock.patch(RAY_STRING.format("vertex_ray.update_ray_cluster"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    @mock.patch(RAY_STRING.format("PersistentResourceServiceClient.persistent_resource_path"))
    def test_update_ray_cluster(
        self, mock_persistent_resource_path, mock_aiplatform_init, mock_update_ray_cluster
    ) -> None:
        self.hook.update_ray_cluster(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            cluster_id=TEST_CLUSTER_NAME,
            worker_node_types=[TEST_NODE_RESOURCES],
        )
        mock_aiplatform_init.assert_called_once()
        mock_persistent_resource_path.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_LOCATION,
            persistent_resource=TEST_CLUSTER_NAME,
        )
        mock_update_ray_cluster.assert_called_once_with(
            cluster_resource_name=mock_persistent_resource_path.return_value,
            worker_node_types=[TEST_NODE_RESOURCES],
        )

    @mock.patch(RAY_STRING.format("vertex_ray.list_ray_clusters"))
    @mock.patch(RAY_STRING.format("aiplatform.init"))
    def test_list_ray_clusters(self, mock_aiplatform_init, mock_list_ray_clusters) -> None:
        self.hook.list_ray_clusters(
            project_id=TEST_PROJECT_ID,
            location=TEST_LOCATION,
        )
        mock_aiplatform_init.assert_called_once()
        mock_list_ray_clusters.assert_called_once()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Calendar API hook."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from googleapiclient.discovery import build

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

if TYPE_CHECKING:
    from datetime import datetime


class GoogleCalendarHook(GoogleBaseHook):
    """
    Interact with Google Calendar via Google Cloud connection.

    Reading and writing cells in Google Sheet: https://developers.google.com/calendar/api/v3/reference

    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param api_version: API Version. For example v3
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        api_version: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
        )
        self.api_version = api_version
        self._conn = None

    def get_conn(self) -> Any:
        """
        Retrieve connection to Google Calendar.

        :return: Google Calendar services object.
        """
        if not self._conn:
            http_authorized = self._authorize()
            self._conn = build("calendar", self.api_version, http=http_authorized, cache_discovery=False)

        return self._conn

    def get_events(
        self,
        calendar_id: str = "primary",
        i_cal_uid: str | None = None,
        max_attendees: int | None = None,
        max_results: int | None = None,
        order_by: str | None = None,
        private_extended_property: str | None = None,
        q: str | None = None,
        shared_extended_property: str | None = None,
        show_deleted: bool | None = False,
        show_hidden_invitation: bool | None = False,
        single_events: bool | None = False,
        sync_token: str | None = None,
        time_max: datetime | None = None,
        time_min: datetime | None = None,
        time_zone: str | None = None,
        updated_min: datetime | None = None,
    ) -> list:
        """
        Get events from Google Calendar from a single calendar_id.

        https://developers.google.com/calendar/api/v3/reference/events/list

        :param calendar_id: The Google Calendar ID to interact with
        :param i_cal_uid: Optional. Specifies event ID in the ``iCalendar`` format in the response.
        :param max_attendees: Optional. If there are more than the specified number of attendees,
            only the participant is returned.
        :param max_results: Optional. Maximum number of events returned on one result page.
            Incomplete pages can be detected by a non-empty ``nextPageToken`` field in the response.
            By default, the value is 250 events. The page size can never be larger than 2500 events
        :param order_by: Optional. Acceptable values are ``"startTime"`` or "updated"
        :param private_extended_property: Optional. Extended properties constraint specified as
            ``propertyName=value``. Matches only private properties. This parameter might be repeated
            multiple times to return events that match all given constraints.
        :param q: Optional. Free text search.
        :param shared_extended_property: Optional. Extended properties constraint specified as
            ``propertyName=value``. Matches only shared properties. This parameter might be repeated
            multiple times to return events that match all given constraints.
        :param show_deleted: Optional. False by default
        :param show_hidden_invitation: Optional. False by default
        :param single_events: Optional. False by default
        :param sync_token: Optional. Token obtained from the ``nextSyncToken`` field returned
        :param time_max: Optional. Upper bound (exclusive) for an event's start time to filter by.
            Default is no filter
        :param time_min: Optional. Lower bound (exclusive) for an event's end time to filter by.
            Default is no filter
        :param time_zone: Optional. Time zone used in response. Default is calendars time zone.
        :param updated_min: Optional. Lower bound for an event's last modification time
        """
        service = self.get_conn()
        page_token = None
        events = []
        while True:
            response = (
                service.events()
                .list(
                    calendarId=calendar_id,
                    iCalUID=i_cal_uid,
                    maxAttendees=max_attendees,
                    maxResults=max_results,
                    orderBy=order_by,
                    pageToken=page_token,
                    privateExtendedProperty=private_extended_property,
                    q=q,
                    sharedExtendedProperty=shared_extended_property,
                    showDeleted=show_deleted,
                    showHiddenInvitations=show_hidden_invitation,
                    singleEvents=single_events,
                    syncToken=sync_token,
                    timeMax=time_max,
                    timeMin=time_min,
                    timeZone=time_zone,
                    updatedMin=updated_min,
                )
                .execute(num_retries=self.num_retries)
            )
            events.extend(response["items"])
            page_token = response.get("nextPageToken")
            if not page_token:
                break
        return events

    def create_event(
        self,
        event: dict[str, Any],
        calendar_id: str = "primary",
        conference_data_version: int | None = 0,
        max_attendees: int | None = None,
        send_notifications: bool | None = False,
        send_updates: str | None = "false",
        supports_attachments: bool | None = False,
    ) -> dict:
        """
        Create event on the specified calendar.

        https://developers.google.com/calendar/api/v3/reference/events/insert.

        :param calendar_id: The Google Calendar ID to interact with
        :param conference_data_version: Optional. Version number of conference data
            supported by the API client.
        :param max_attendees: Optional. If there are more than the specified number of attendees,
            only the participant is returned.
        :param send_notifications: Optional. Default is False
        :param send_updates: Optional. Default is "false". Acceptable values as "all", "none",
            ``"externalOnly"``
            https://developers.google.com/calendar/api/v3/reference/events#resource
        """
        if "start" not in event or "end" not in event:
            raise AirflowException(
                f"start and end must be specified in the event body while creating an event. API docs:"
                f"https://developers.google.com/calendar/api/{self.api_version}/reference/events/insert "
            )
        service = self.get_conn()

        response = (
            service.events()
            .insert(
                calendarId=calendar_id,
                conferenceDataVersion=conference_data_version,
                maxAttendees=max_attendees,
                sendNotifications=send_notifications,
                sendUpdates=send_updates,
                supportsAttachments=supports_attachments,
                body=event,
            )
            .execute(num_retries=self.num_retries)
        )

        return response



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from pathlib import Path
from unittest import mock

from airflow.providers.google.suite.transfers.local_to_drive import LocalFilesystemToGoogleDriveOperator

GCP_CONN_ID = "test"
DRIVE_FOLDER = Path("test_folder")
LOCAL_PATHS = [Path("test1"), Path("test2")]
REMOTE_FILE_IDS = ["rtest1", "rtest2"]


class TestLocalFilesystemToGoogleDriveOperator:
    @mock.patch("airflow.providers.google.suite.transfers.local_to_drive.GoogleDriveHook")
    def test_execute(self, mock_hook):
        context = {}
        mock_hook.return_value.upload_file.return_value = REMOTE_FILE_IDS
        op = LocalFilesystemToGoogleDriveOperator(
            task_id="test_task",
            local_paths=LOCAL_PATHS,
            drive_folder=DRIVE_FOLDER,
            gcp_conn_id=GCP_CONN_ID,
            folder_id="some_folder_id",
        )
        op.execute(context)

        calls = [
            mock.call(
                local_location="test1",
                remote_location="test_folder/test1",
                chunk_size=100 * 1024 * 1024,
                resumable=False,
                folder_id="some_folder_id",
                show_full_target_path=True,
            ),
            mock.call(
                local_location="test2",
                remote_location="test_folder/test2",
                chunk_size=100 * 1024 * 1024,
                resumable=False,
                folder_id="some_folder_id",
                show_full_target_path=True,
            ),
        ]
        mock_hook.return_value.upload_file.assert_has_calls(calls)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import Any

from requests.exceptions import HTTPError

from tests_common.test_utils.api_client_helpers import make_authenticated_rest_api_request


def create_connection_request(connection_id: str, connection: dict[str, Any], is_composer: bool = False):
    if is_composer:
        from airflow.providers.common.compat.sdk import conf
        from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerHook

        hook = CloudComposerHook()
        composer_airflow_uri = conf.get("api", "base_url").rstrip("/")

        response = hook.make_composer_airflow_api_request(
            method="POST",
            airflow_uri=composer_airflow_uri,
            path="/api/v2/connections",
            data=json.dumps(
                {
                    "connection_id": connection_id,
                    **connection,
                }
            ),
        )
        response.raise_for_status()
        if response.text != "":
            return response.json()
    else:
        return make_authenticated_rest_api_request(
            path="/api/v2/connections",
            method="POST",
            body={
                "connection_id": connection_id,
                **connection,
            },
        )


def delete_connection_request(connection_id: str, is_composer: bool = False):
    if is_composer:
        from airflow.providers.common.compat.sdk import conf
        from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerHook

        hook = CloudComposerHook()
        composer_airflow_uri = conf.get("api", "base_url").rstrip("/")

        response = hook.make_composer_airflow_api_request(
            method="DELETE", airflow_uri=composer_airflow_uri, path=f"/api/v2/connections/{connection_id}"
        )
        response.raise_for_status()
        if response.text != "":
            return response.json()
    else:
        return make_authenticated_rest_api_request(
            path=f"/api/v2/connections/{connection_id}",
            method="DELETE",
        )


def create_airflow_connection(
    connection_id: str, connection_conf: dict[str, Any], is_composer: bool = False
) -> None:
    from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

    print(f"Removing connection '{connection_id}' if it exists")
    if AIRFLOW_V_3_0_PLUS:
        try:
            delete_connection_request(connection_id=connection_id, is_composer=is_composer)
        except HTTPError:
            print(f"Connection '{connection_id}' does not exist. A new one will be created")
        create_connection_request(
            connection_id=connection_id, connection=connection_conf, is_composer=is_composer
        )
    else:
        from sqlalchemy import delete

        from airflow.models import Connection
        from airflow.settings import Session

        if Session is None:
            raise RuntimeError("Session not configured. Call configure_orm() first.")
        session = Session()
        session.execute(delete(Connection).where(Connection.conn_id == connection_id))
        connection = Connection(conn_id=connection_id, **connection_conf)
        session.add(connection)
        session.commit()
    print(f"Connection '{connection_id}' created")


def delete_airflow_connection(connection_id: str, is_composer: bool = False) -> None:
    from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

    print(f"Removing connection '{connection_id}'")
    if AIRFLOW_V_3_0_PLUS:
        delete_connection_request(connection_id=connection_id, is_composer=is_composer)
    else:
        from sqlalchemy import delete

        from airflow.models import Connection
        from airflow.settings import Session

        if Session is None:
            raise RuntimeError("Session not configured. Call configure_orm() first.")
        session = Session()
        session.execute(delete(Connection).where(Connection.conn_id == connection_id))
        session.commit()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
import json
import logging
import os
from unittest.mock import MagicMock, patch

import pytest
from google.cloud.bigquery.table import Table

from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    Dataset,
    DocumentationDatasetFacet,
    ExternalQueryRunFacet,
    Fields,
    InputDataset,
    InputField,
    OutputDataset,
    OutputStatisticsOutputDatasetFacet,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
)
from airflow.providers.google.cloud.openlineage.facets import BigQueryJobRunFacet
from airflow.providers.google.cloud.openlineage.mixins import _BigQueryInsertJobOperatorOpenLineageMixin
from airflow.providers.openlineage.sqlparser import SQLParser

QUERY_JOB_PROPERTIES = {
    "configuration": {
        "query": {
            "query": """
            INSERT INTO dest_project.dest_dataset.dest_table
            SELECT a, b, c FROM source_project.source_dataset.source_table
            UNION ALL
            SELECT a, b, c FROM source_table2
            """,
            "defaultDataset": {"datasetId": "default_dataset", "projectId": "default_project"},
        }
    }
}
OUTPUT_DATASET = OutputDataset(
    namespace="bigquery",
    name="dest_project.dest_dataset.dest_table",
    facets={
        "schema": SchemaDatasetFacet(
            fields=[
                SchemaDatasetFacetFields("a", "STRING"),
                SchemaDatasetFacetFields("b", "STRING"),
                SchemaDatasetFacetFields("c", "STRING"),
                SchemaDatasetFacetFields("d", "STRING"),
                SchemaDatasetFacetFields("e", "STRING"),
                SchemaDatasetFacetFields("f", "STRING"),
                SchemaDatasetFacetFields("g", "STRING"),
            ]
        )
    },
)
INPUT_DATASETS = [
    InputDataset(
        namespace="bigquery",
        name="source_project.source_dataset.source_table",
        facets={
            "schema": SchemaDatasetFacet(
                fields=[
                    SchemaDatasetFacetFields("a", "STRING"),
                    SchemaDatasetFacetFields("b", "STRING"),
                    SchemaDatasetFacetFields("c", "STRING"),
                    SchemaDatasetFacetFields("x", "STRING"),
                ]
            )
        },
    ),
    InputDataset(
        namespace="bigquery",
        name="default_project.default_dataset.source_table2",
        facets={
            "schema": SchemaDatasetFacet(
                fields=[
                    SchemaDatasetFacetFields("a", "STRING"),
                    SchemaDatasetFacetFields("b", "STRING"),
                    SchemaDatasetFacetFields("c", "STRING"),
                    SchemaDatasetFacetFields("y", "STRING"),
                ]
            )
        },
    ),
    InputDataset("bigquery", "some.random.tb"),
]


def read_common_json_file(rel: str):
    with open(os.path.dirname(__file__) + "/../utils/" + rel) as f:
        return json.load(f)


class TestBigQueryOpenLineageMixin:
    def setup_method(self):
        self.copy_job_details = read_common_json_file("copy_job_details.json")
        self.extract_job_details = read_common_json_file("extract_job_details.json")
        self.load_job_details = read_common_json_file("load_job_details.json")
        self.query_job_details = read_common_json_file("query_job_details.json")
        self.script_job_details = read_common_json_file("script_job_details.json")
        hook = MagicMock()
        self.client = MagicMock()

        class BQOperator(_BigQueryInsertJobOperatorOpenLineageMixin):
            sql = ""
            job_id = "job_id"
            project_id = "project_id"
            location = None
            log = logging.getLogger("BQOperator")

            @property
            def hook(self):
                return hook

        hook.get_client.return_value = self.client

        self.operator = BQOperator()
        self.operator._client = self.client

    def test_get_openlineage_facets_on_complete_query_job(self):
        self.client.get_job.return_value._properties = self.query_job_details
        self.client.get_table.side_effect = [
            Table.from_api_repr(read_common_json_file("table_details.json")),
            Table.from_api_repr(read_common_json_file("out_table_details.json")),
        ]

        lineage = self.operator.get_openlineage_facets_on_complete(None)

        self.query_job_details["configuration"]["query"].pop("query")
        assert lineage.run_facets == {
            "bigQueryJob": BigQueryJobRunFacet(
                cached=False, billedBytes=111149056, properties=json.dumps(self.query_job_details)
            ),
            "externalQuery": ExternalQueryRunFacet(externalQueryId="job_id", source="bigquery"),
        }
        assert lineage.inputs == [
            InputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.test_table",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("state", "STRING", "2-digit state code"),
                            SchemaDatasetFacetFields("gender", "STRING", "Sex (M=male or F=female)"),
                            SchemaDatasetFacetFields("year", "INTEGER", "4-digit year of birth"),
                            SchemaDatasetFacetFields("name", "STRING", "Given name of a person at birth"),
                            SchemaDatasetFacetFields(
                                "number", "INTEGER", "Number of occurrences of the name"
                            ),
                        ]
                    ),
                    "documentation": DocumentationDatasetFacet(
                        "The table contains the number of applicants for a Social Security card by year of birth and sex."
                    ),
                },
            )
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.output_table",
                outputFacets={
                    "outputStatistics": OutputStatisticsOutputDatasetFacet(
                        rowCount=20, size=321, fileCount=None
                    )
                },
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    ),
                },
            ),
        ]

    def test_get_openlineage_facets_on_complete_copy_job(self):
        self.client.get_job.return_value._properties = self.copy_job_details
        self.client.get_table.return_value = Table.from_api_repr(read_common_json_file("table_details.json"))
        expected_facets = {
            "schema": SchemaDatasetFacet(
                fields=[
                    SchemaDatasetFacetFields("state", "STRING", "2-digit state code"),
                    SchemaDatasetFacetFields("gender", "STRING", "Sex (M=male or F=female)"),
                    SchemaDatasetFacetFields("year", "INTEGER", "4-digit year of birth"),
                    SchemaDatasetFacetFields("name", "STRING", "Given name of a person at birth"),
                    SchemaDatasetFacetFields("number", "INTEGER", "Number of occurrences of the name"),
                ]
            ),
            "documentation": DocumentationDatasetFacet(
                "The table contains the number of applicants for a Social Security card by year of birth and sex."
            ),
        }

        lineage = self.operator.get_openlineage_facets_on_complete(None)

        assert lineage.run_facets == {
            "bigQueryJob": BigQueryJobRunFacet(
                cached=False, billedBytes=None, properties=json.dumps(self.copy_job_details)
            ),
            "externalQuery": ExternalQueryRunFacet(externalQueryId="job_id", source="bigquery"),
        }
        assert lineage.inputs == [
            InputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.copy_job_source",
                facets=expected_facets,
            ),
            InputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.copy_job_source2",
                facets=expected_facets,
            ),
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.copy_job_result",
                outputFacets={
                    "outputStatistics": OutputStatisticsOutputDatasetFacet(
                        rowCount=20, size=3800, fileCount=None
                    )
                },
                facets={
                    **expected_facets,
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            col: Fields(
                                inputFields=[
                                    InputField(
                                        "bigquery", "airflow-openlineage.new_dataset.copy_job_source", col
                                    ),
                                    InputField(
                                        "bigquery", "airflow-openlineage.new_dataset.copy_job_source2", col
                                    ),
                                ],
                                transformationType="IDENTITY",
                                transformationDescription="identical",
                            )
                            for col in ["state", "gender", "year", "name", "number"]
                        }
                    ),
                },
            ),
        ]

    def test_get_openlineage_facets_on_complete_load_job(self):
        self.client.get_job.return_value._properties = self.load_job_details
        self.client.get_table.return_value = Table.from_api_repr(
            read_common_json_file("out_table_details.json")
        )

        lineage = self.operator.get_openlineage_facets_on_complete(None)

        assert lineage.run_facets == {
            "bigQueryJob": BigQueryJobRunFacet(
                cached=False, billedBytes=None, properties=json.dumps(self.load_job_details)
            ),
            "externalQuery": ExternalQueryRunFacet(externalQueryId="job_id", source="bigquery"),
        }
        assert lineage.inputs == [
            InputDataset(
                namespace="gs://airflow-openlineage",
                name="/",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    )
                },
            ),
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.job_load",
                outputFacets={
                    "outputStatistics": OutputStatisticsOutputDatasetFacet(
                        rowCount=10, size=546, fileCount=None
                    )
                },
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    ),
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            "name": Fields(
                                inputFields=[InputField("gs://airflow-openlineage", "/", "name")],
                                transformationType="IDENTITY",
                                transformationDescription="identical",
                            ),
                            "total_people": Fields(
                                inputFields=[InputField("gs://airflow-openlineage", "/", "total_people")],
                                transformationType="IDENTITY",
                                transformationDescription="identical",
                            ),
                        }
                    ),
                },
            ),
        ]

    def test_get_openlineage_facets_on_complete_extract_job(self):
        self.client.get_job.return_value._properties = self.extract_job_details
        self.client.get_table.return_value = Table.from_api_repr(
            read_common_json_file("out_table_details.json")
        )

        lineage = self.operator.get_openlineage_facets_on_complete(None)

        assert lineage.run_facets == {
            "bigQueryJob": BigQueryJobRunFacet(
                cached=False, billedBytes=None, properties=json.dumps(self.extract_job_details)
            ),
            "externalQuery": ExternalQueryRunFacet(externalQueryId="job_id", source="bigquery"),
        }
        assert lineage.inputs == [
            InputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.extract_job_source",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    )
                },
            ),
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="gs://airflow-openlineage",
                name="extract_job_source",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    ),
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            "name": Fields(
                                inputFields=[
                                    InputField(
                                        "bigquery",
                                        "airflow-openlineage.new_dataset.extract_job_source",
                                        "name",
                                    )
                                ],
                                transformationType="IDENTITY",
                                transformationDescription="identical",
                            ),
                            "total_people": Fields(
                                inputFields=[
                                    InputField(
                                        "bigquery",
                                        "airflow-openlineage.new_dataset.extract_job_source",
                                        "total_people",
                                    )
                                ],
                                transformationType="IDENTITY",
                                transformationDescription="identical",
                            ),
                        }
                    ),
                },
            ),
        ]

    def test_get_openlineage_facets_on_complete_script_job(self):
        self.client.get_job.side_effect = [
            MagicMock(_properties=self.script_job_details),
            MagicMock(_properties=self.query_job_details),
        ]
        self.client.get_table.side_effect = [
            Table.from_api_repr(read_common_json_file("table_details.json")),
            Table.from_api_repr(read_common_json_file("out_table_details.json")),
        ]
        self.client.list_jobs.return_value = ["child_job_id"]

        lineage = self.operator.get_openlineage_facets_on_complete(None)

        self.script_job_details["configuration"]["query"].pop("query")
        assert lineage.run_facets == {
            "bigQueryJob": BigQueryJobRunFacet(
                cached=False, billedBytes=120586240, properties=json.dumps(self.script_job_details)
            ),
            "externalQuery": ExternalQueryRunFacet(externalQueryId="job_id", source="bigquery"),
        }
        assert lineage.inputs == [
            InputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.test_table",
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("state", "STRING", "2-digit state code"),
                            SchemaDatasetFacetFields("gender", "STRING", "Sex (M=male or F=female)"),
                            SchemaDatasetFacetFields("year", "INTEGER", "4-digit year of birth"),
                            SchemaDatasetFacetFields("name", "STRING", "Given name of a person at birth"),
                            SchemaDatasetFacetFields(
                                "number", "INTEGER", "Number of occurrences of the name"
                            ),
                        ]
                    ),
                    "documentation": DocumentationDatasetFacet(
                        "The table contains the number of applicants for a Social Security card by year of birth and sex."
                    ),
                },
            )
        ]
        assert lineage.outputs == [
            OutputDataset(
                namespace="bigquery",
                name="airflow-openlineage.new_dataset.output_table",
                outputFacets={
                    "outputStatistics": OutputStatisticsOutputDatasetFacet(
                        rowCount=20, size=321, fileCount=None
                    )
                },
                facets={
                    "schema": SchemaDatasetFacet(
                        fields=[
                            SchemaDatasetFacetFields("name", "STRING"),
                            SchemaDatasetFacetFields("total_people", "INTEGER"),
                        ]
                    ),
                },
            ),
        ]

    def test_deduplicate_outputs(self):
        outputs = [
            None,
            OutputDataset(
                name="d1",
                namespace="",
                outputFacets={"outputStatistics": OutputStatisticsOutputDatasetFacet(3, 4)},
            ),
            OutputDataset(
                name="d1",
                namespace="",
                outputFacets={"outputStatistics": OutputStatisticsOutputDatasetFacet(3, 4)},
                facets={"t1": "t1"},
            ),
            OutputDataset(
                name="d2",
                namespace="",
                outputFacets={"outputStatistics": OutputStatisticsOutputDatasetFacet(6, 7)},
                facets={"t2": "t2"},
            ),
            OutputDataset(
                name="d2",
                namespace="",
                outputFacets={"outputStatistics": OutputStatisticsOutputDatasetFacet(60, 70)},
                facets={"t20": "t20"},
            ),
        ]
        result = self.operator._deduplicate_outputs(outputs)
        assert len(result) == 2
        first_result = result[0]
        assert first_result.name == "d1"
        assert first_result.facets == {"t1": "t1"}
        second_result = result[1]
        assert second_result.name == "d2"
        assert second_result.facets == {"t20": "t20"}

    def test_deduplicate_outputs_with_cll(self):
        outputs = [
            None,
            OutputDataset(
                name="a.b.c",
                namespace="bigquery",
                facets={
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            "c": Fields(
                                inputFields=[InputField("bigquery", "a.b.1", "c")],
                                transformationType="",
                                transformationDescription="",
                            ),
                            "d": Fields(
                                inputFields=[InputField("bigquery", "a.b.2", "d")],
                                transformationType="",
                                transformationDescription="",
                            ),
                        }
                    )
                },
            ),
            OutputDataset(
                name="a.b.c",
                namespace="bigquery",
                facets={
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            "c": Fields(
                                inputFields=[InputField("bigquery", "a.b.3", "x")],
                                transformationType="",
                                transformationDescription="",
                            ),
                            "e": Fields(
                                inputFields=[InputField("bigquery", "a.b.1", "e")],
                                transformationType="",
                                transformationDescription="",
                            ),
                        }
                    )
                },
            ),
            OutputDataset(
                name="x.y.z",
                namespace="bigquery",
                facets={
                    "columnLineage": ColumnLineageDatasetFacet(
                        fields={
                            "c": Fields(
                                inputFields=[InputField("bigquery", "a.b.3", "x")],
                                transformationType="",
                                transformationDescription="",
                            )
                        }
                    )
                },
            ),
        ]
        result = self.operator._deduplicate_outputs(outputs)
        assert len(result) == 2
        first_result = result[0]
        assert first_result.name == "a.b.c"
        assert first_result.facets["columnLineage"] == ColumnLineageDatasetFacet(
            fields={
                "c": Fields(
                    inputFields=[InputField("bigquery", "a.b.1", "c"), InputField("bigquery", "a.b.3", "x")],
                    transformationType="",
                    transformationDescription="",
                ),
                "d": Fields(
                    inputFields=[InputField("bigquery", "a.b.2", "d")],
                    transformationType="",
                    transformationDescription="",
                ),
                "e": Fields(
                    inputFields=[InputField("bigquery", "a.b.1", "e")],
                    transformationType="",
                    transformationDescription="",
                ),
            }
        )
        second_result = result[1]
        assert second_result.name == "x.y.z"
        assert second_result.facets["columnLineage"] == ColumnLineageDatasetFacet(
            fields={
                "c": Fields(
                    inputFields=[InputField("bigquery", "a.b.3", "x")],
                    transformationType="",
                    transformationDescription="",
                )
            }
        )

    @patch("airflow.providers.google.cloud.openlineage.mixins.get_facets_from_bq_table")
    def test_get_table_facets_safely(self, mock_get_facets):
        mock_get_facets.return_value = {"some": "facets"}
        result = self.operator._get_table_facets_safely("some_table")
        assert result == {"some": "facets"}

    @patch("airflow.providers.google.cloud.openlineage.mixins.get_facets_from_bq_table")
    def test_get_table_facets_safely_empty_when_error(self, mock_get_facets):
        mock_get_facets.side_effect = ValueError("Some error")
        result = self.operator._get_table_facets_safely("some_table")
        assert result == {}

    @patch("airflow.providers.google.cloud.openlineage.mixins.get_facets_from_bq_table")
    def test_get_dataset(self, mock_get_facets):
        mock_get_facets.return_value = {"some": "facets"}
        table_ref = {"projectId": "p", "datasetId": "d", "tableId": "t"}
        input_result = self.operator._get_dataset(table_ref, "input")
        assert input_result == InputDataset("bigquery", "p.d.t", facets={"some": "facets"})

        output_result = self.operator._get_dataset(table_ref, "output")
        assert output_result == OutputDataset("bigquery", "p.d.t", facets={"some": "facets"})

        with pytest.raises(ValueError, match="Invalid dataset_type. Must be 'input' or 'output'"):
            self.operator._get_dataset(table_ref, "wrong")

    @patch("airflow.providers.google.cloud.openlineage.mixins.get_facets_from_bq_table")
    def test_get_input_dataset(self, mock_get_facets):
        mock_get_facets.return_value = {"some": "facets"}
        table_ref = {"projectId": "p", "datasetId": "d", "tableId": "t"}
        expected_result = self.operator._get_dataset(table_ref, "input")
        result = self.operator._get_input_dataset(table_ref)
        assert result == expected_result

    @patch("airflow.providers.google.cloud.openlineage.mixins.get_facets_from_bq_table")
    def test_get_output_dataset(self, mock_get_facets):
        mock_get_facets.return_value = {"some": "facets"}
        table_ref = {"projectId": "p", "datasetId": "d", "tableId": "t"}
        expected_result = self.operator._get_dataset(table_ref, "output")
        result = self.operator._get_output_dataset(table_ref)
        assert result == expected_result

    @pytest.mark.parametrize("job_type", ("LOAD", "COPY", "EXTRACT"))
    def test_get_bigquery_job_run_facet_non_query_jobs(self, job_type):
        properties = {
            "statistics": {"some": "stats"},
            "configuration": {"jobType": job_type},
        }
        result = self.operator._get_bigquery_job_run_facet(properties)
        assert result.cached is False
        assert result.billedBytes is None
        assert result.properties == json.dumps(properties)

    @pytest.mark.parametrize("cache", (None, "false", False, 0))
    def test_get_bigquery_job_run_facet_query_no_cache_and_with_bytes(self, cache):
        properties = {
            "statistics": {"query": {"cacheHit": cache, "totalBytesBilled": 10}},
            "configuration": {"query": {"query": "SELECT ..."}, "jobType": "QUERY"},
        }
        result = self.operator._get_bigquery_job_run_facet(properties)
        assert result.cached is False
        assert result.billedBytes == 10
        properties["configuration"]["query"].pop("query")
        assert result.properties == json.dumps(properties)

    @pytest.mark.parametrize("cache", ("true", True))
    def test_get_bigquery_job_run_facet_query_with_cache_and_no_bytes(self, cache):
        properties = {
            "statistics": {
                "query": {
                    "cacheHit": cache,
                }
            },
            "configuration": {"query": {"query": "SELECT ..."}, "jobType": "QUERY"},
        }
        result = self.operator._get_bigquery_job_run_facet(properties)
        assert result.cached is True
        assert result.billedBytes is None
        properties["configuration"]["query"].pop("query")
        assert result.properties == json.dumps(properties)

    def test_get_output_statistics_dataset_facet_query_no_query_plan(self):
        properties = {
            "statistics": {"query": {"totalBytesBilled": 10}},
            "configuration": {"query": {"query": "SELECT ..."}, "jobType": "QUERY"},
        }
        result = self.operator._get_output_statistics_dataset_facet(properties)
        assert result is None

    def test_get_output_statistics_dataset_facet_query_no_stats(self):
        properties = {
            "statistics": {"query": {"totalBytesBilled": 10, "queryPlan": [{"test": "test"}]}},
            "configuration": {"query": {"query": "SELECT ..."}, "jobType": "QUERY"},
        }
        result = self.operator._get_output_statistics_dataset_facet(properties)
        assert result is None

    def test_get_output_statistics_dataset_facet_query(self):
        properties = {
            "statistics": {
                "query": {
                    "totalBytesBilled": 10,
                    "queryPlan": [{"recordsWritten": 123, "shuffleOutputBytes": "321"}],
                }
            },
            "configuration": {"query": {"query": "SELECT ..."}, "jobType": "QUERY"},
        }
        result = self.operator._get_output_statistics_dataset_facet(properties)
        assert result.rowCount == 123
        assert result.size == 321

    def test_get_output_statistics_dataset_facet_copy(self):
        properties = {
            "statistics": {
                "copy": {
                    "copiedRows": 123,
                    "copiedLogicalBytes": 321,
                }
            },
            "configuration": {"jobType": "COPY"},
        }
        result = self.operator._get_output_statistics_dataset_facet(properties)
        assert result.rowCount == 123
        assert result.size == 321

    def test_get_output_statistics_dataset_facet_load(self):
        properties = {
            "statistics": {
                "load": {
                    "outputRows": 123,
                    "outputBytes": 321,
                }
            },
            "configuration": {"jobType": "LOAD"},
        }
        result = self.operator._get_output_statistics_dataset_facet(properties)
        assert result.rowCount == 123
        assert result.size == 321

    def test_get_column_level_lineage_facet(self):
        result = self.operator._get_column_level_lineage_facet_for_query_job(
            QUERY_JOB_PROPERTIES, OUTPUT_DATASET, INPUT_DATASETS
        )
        assert result == ColumnLineageDatasetFacet(
            fields={
                col: Fields(
                    inputFields=[
                        InputField("bigquery", "default_project.default_dataset.source_table2", col),
                        InputField("bigquery", "source_project.source_dataset.source_table", col),
                    ],
                    transformationType="",
                    transformationDescription="",
                )
                for col in ("a", "b", "c")
            }
        )

    def test_get_column_level_lineage_facet_early_exit_empty_cll_from_parser(self):
        properties = {"configuration": {"query": {"query": "SELECT 1"}}}
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job(
                properties, OUTPUT_DATASET, INPUT_DATASETS
            )
            is None
        )
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job({}, OUTPUT_DATASET, INPUT_DATASETS)
            is None
        )

    def test_get_column_level_lineage_facet_early_exit_output_table_id_mismatch(self):
        output = copy.deepcopy(OUTPUT_DATASET)
        output.name = "different.name.table"
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job(
                QUERY_JOB_PROPERTIES, output, INPUT_DATASETS
            )
            is None
        )

    def test_get_column_level_lineage_facet_early_exit_output_columns_mismatch(self):
        output = copy.deepcopy(OUTPUT_DATASET)
        output.facets["schema"].fields = [
            SchemaDatasetFacetFields("different_col", "STRING"),
        ]
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job(
                QUERY_JOB_PROPERTIES, output, INPUT_DATASETS
            )
            is None
        )

    def test_get_column_level_lineage_facet_early_exit_wrong_parsed_input_tables(self):
        properties = {
            "configuration": {
                "query": {
                    "query": """
                    INSERT INTO dest_project.dest_dataset.dest_table
                    SELECT a, b, c FROM some.wrong.source_table
                    """,
                }
            }
        }
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job(
                properties, OUTPUT_DATASET, INPUT_DATASETS
            )
            is None
        )

    def test_get_column_level_lineage_facet_early_exit_wrong_parsed_input_columns(self):
        properties = {
            "configuration": {
                "query": {
                    "query": """
                    INSERT INTO dest_project.dest_dataset.dest_table
                    SELECT wrong_col, wrong2, wrong3 FROM source_project.source_dataset.source_table
                    """,
                }
            }
        }
        assert (
            self.operator._get_column_level_lineage_facet_for_query_job(
                properties, OUTPUT_DATASET, INPUT_DATASETS
            )
            is None
        )

    def test_get_qualified_name_from_parse_result(self):
        class _Table:  # Replacement for SQL parser TableMeta
            database = "project"
            schema = "dataset"
            name = "table"

        class _TableNoSchema:  # Replacement for SQL parser TableMeta
            database = None
            schema = "dataset"
            name = "table"

        class _TableNoSchemaNoDb:  # Replacement for SQL parser TableMeta
            database = None
            schema = None
            name = "table"

        result = self.operator._get_qualified_name_from_parse_result(
            table=_Table(),
            default_project="default_project",
            default_dataset="default_dataset",
        )
        assert result == "project.dataset.table"

        result = self.operator._get_qualified_name_from_parse_result(
            table=_TableNoSchema(),
            default_project="default_project",
            default_dataset="default_dataset",
        )
        assert result == "default_project.dataset.table"

        result = self.operator._get_qualified_name_from_parse_result(
            table=_TableNoSchemaNoDb(),
            default_project="default_project",
            default_dataset="default_dataset",
        )
        assert result == "default_project.default_dataset.table"

    def test_extract_default_dataset_and_project(self):
        properties = {"configuration": {"query": {"defaultDataset": {"datasetId": "default_dataset"}}}}
        result = self.operator._extract_default_dataset_and_project(properties, "default_project")
        assert result == ("default_dataset", "default_project")

        properties = {
            "configuration": {
                "query": {"defaultDataset": {"datasetId": "default_dataset", "projectId": "default_project"}}
            }
        }
        result = self.operator._extract_default_dataset_and_project(properties, "another_project")
        assert result == ("default_dataset", "default_project")

        result = self.operator._extract_default_dataset_and_project({}, "default_project")
        assert result == ("", "default_project")

    def test_validate_output_table_id_no_table(self):
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("SELECT 1"))
        assert parse_result.out_tables == []
        assert self.operator._validate_output_table_id(parse_result, None, None, None) is False

    def test_validate_output_table_id_multiple_tables(self):
        query = "INSERT INTO a.b.c VALUES (1); INSERT INTO d.e.f VALUES (2);"
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string(query))
        assert len(parse_result.out_tables) == 2
        assert self.operator._validate_output_table_id(parse_result, None, None, None) is False

    def test_validate_output_table_id_mismatch(self):
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("INSERT INTO a.b.c VALUES (1)"))
        assert len(parse_result.out_tables) == 1
        assert parse_result.out_tables[0].qualified_name == "a.b.c"
        assert (
            self.operator._validate_output_table_id(parse_result, OutputDataset("", "d.e.f"), None, None)
            is False
        )

    def test_validate_output_table_id(self):
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("INSERT INTO a.b.c VALUES (1)"))
        assert len(parse_result.out_tables) == 1
        assert parse_result.out_tables[0].qualified_name == "a.b.c"
        assert (
            self.operator._validate_output_table_id(parse_result, OutputDataset("", "a.b.c"), None, None)
            is True
        )

    def test_validate_output_table_id_query_with_table_name_only(self):
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("INSERT INTO c VALUES (1)"))
        assert len(parse_result.out_tables) == 1
        assert parse_result.out_tables[0].qualified_name == "c"
        assert (
            self.operator._validate_output_table_id(parse_result, OutputDataset("", "a.b.c"), "a", "b")
            is True
        )

    def test_extract_column_names_dataset_without_schema(self):
        assert self.operator._extract_column_names(Dataset("a", "b")) == []

    def test_extract_column_names_dataset_(self):
        ds = Dataset(
            "a",
            "b",
            facets={
                "schema": SchemaDatasetFacet(
                    fields=[
                        SchemaDatasetFacetFields("col1", "STRING"),
                        SchemaDatasetFacetFields("col2", "STRING"),
                    ]
                )
            },
        )
        assert self.operator._extract_column_names(ds) == ["col1", "col2"]

    def test_validate_output_columns_mismatch(self):
        ds = OutputDataset(
            "a",
            "b",
            facets={
                "schema": SchemaDatasetFacet(
                    fields=[
                        SchemaDatasetFacetFields("col1", "STRING"),
                        SchemaDatasetFacetFields("col2", "STRING"),
                    ]
                )
            },
        )
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("SELECT a , b FROM c"))
        assert self.operator._validate_output_columns(parse_result, ds) is False

    def test_validate_output_columns(self):
        ds = OutputDataset(
            "a",
            "b",
            facets={
                "schema": SchemaDatasetFacet(
                    fields=[
                        SchemaDatasetFacetFields("a", "STRING"),
                        SchemaDatasetFacetFields("b", "STRING"),
                    ]
                )
            },
        )
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("SELECT a , b FROM c"))
        assert self.operator._validate_output_columns(parse_result, ds) is True

    def test_extract_parsed_input_tables(self):
        query = "INSERT INTO x SELECT a, b from project1.ds1.tb1; INSERT INTO y SELECT c, d from tb2;"
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string(query))
        assert self.operator._extract_parsed_input_tables(parse_result, "default_project", "default_ds") == {
            "project1.ds1.tb1": ["a", "b"],
            "default_project.default_ds.tb2": ["c", "d"],
        }

    def test_extract_parsed_input_tables_no_cll(self):
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string("SELECT 1"))
        assert self.operator._extract_parsed_input_tables(parse_result, "p", "d") == {}

    def test_validate_input_tables_mismatch(self):
        result = self.operator._validate_input_tables({"a": None, "b": None}, {"a": None, "c": None})
        assert result is False

    def test_validate_input_tables_bq_has_more_tables(self):
        result = self.operator._validate_input_tables({"a": None}, {"a": None, "c": None})
        assert result is True

    def test_validate_input_tables_empty(self):
        result = self.operator._validate_input_tables({}, {"a": None, "c": None})
        assert result is False

    def test_validate_input_columns_mismatch(self):
        result = self.operator._validate_input_columns(
            {"a": ["1", "2"], "b": ["3", "4"]}, {"a": ["1", "2", "3"], "c": ["4", "5"]}
        )
        assert result is False

    def test_validate_input_columns_bq_has_more_cols(self):
        result = self.operator._validate_input_columns(
            {"a": ["1", "2"]}, {"a": ["1", "2", "3"], "c": ["4", "5"]}
        )
        assert result is True

    def test_validate_input_columns_empty(self):
        result = self.operator._validate_input_columns({}, {"a": ["1", "2", "3"], "c": ["4", "5"]})
        assert result is False

    def test_generate_column_lineage_facet(self):
        query = "INSERT INTO b.c SELECT c, d from tb2;"
        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string(query))
        result = self.operator._generate_column_lineage_facet(parse_result, "default_project", "default_ds")
        assert result == ColumnLineageDatasetFacet(
            fields={
                "c": Fields(
                    inputFields=[InputField("bigquery", "default_project.default_ds.tb2", "c")],
                    transformationType="",
                    transformationDescription="",
                ),
                "d": Fields(
                    inputFields=[InputField("bigquery", "default_project.default_ds.tb2", "d")],
                    transformationType="",
                    transformationDescription="",
                ),
            }
        )

    def test_project_id_selection(self):
        """
        Check if project_id set via an argument to the operator takes prevalence over project_id set in a
        connection.
        """
        from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

        class TestOperator(GoogleCloudBaseOperator, _BigQueryInsertJobOperatorOpenLineageMixin):
            def __init__(self, project_id: str | None = None, **_):
                self.project_id = project_id
                self.job_id = "foobar"
                self.location = "foobar"
                self.sql = "foobar"

        # First test task where project_id is set explicitly
        test = TestOperator(project_id="project_a")
        test.hook = MagicMock()
        test.hook.project_id = "project_b"
        test._client = MagicMock()

        test.get_openlineage_facets_on_complete(None)
        _, kwargs = test.hook.get_client.call_args
        assert kwargs["project_id"] == "project_a"

        # Then test task where project_id is inherited from the hook
        test = TestOperator()
        test.hook = MagicMock()
        test.hook.project_id = "project_b"
        test._client = MagicMock()

        test.get_openlineage_facets_on_complete(None)
        _, kwargs = test.hook.get_client.call_args
        assert kwargs["project_id"] == "project_b"
